'' =================================================================================================
''
''   File....... jm_pulse_measure.spin
''   Purpose.... Long-duration pulse measuring
''   Author..... Jon McPhalen
''               Copyright (c) 2024 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 15 MAY 2024
''
'' =================================================================================================


con { timing }

  _xinfreq = 5_000_000                                          ' use 5MHz crystal
  _clkmode = xtal1 + pll16x                                     ' x16 = 80MHz

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate


con { fixed io pins }

  RX1      = 31  { I }                                          ' serial / programming
  TX1      = 30  { O }

  SDA1     = 29  { I/O }                                        ' i2c / eeprom
  SCL1     = 28  { I/O }


con { app io pins }

  PLS_IN   =  0  { I }                                          ' pulse input pin


con

  #true,  ON, OFF
  #false, NO, YES


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz)
  prng : "jm_prng"                                              '   pseudo-random number generation
  io   : "jm_io"                                                '   simple io
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal

' * uses cog when loaded


dat { pre-initialized }


var { globals }


pub main | pw

  setup
  wait_for_terminal(true, 100)

  term.str(string("Long Pulse Measurement", 13, 13))

  repeat
    io.high(PLS_IN)
    time.pause(1000)
    io.low(PLS_IN)

    term.fstr1(string("Measured pulse: %dms\r"), pulsems)


  repeat
    waitcnt(0)


pub setup

'' Setup IO and objects for application

  time.start                                                    ' setup timing & delays

  prng.seed(-cnt, -cnt ~> 2, $CAFE_BABE, cnt <- 2, cnt)         ' seed randomizer

  io.start(0, 0)                                                ' clear all pins (master cog)

  term.tstart(BR_TERM)                                          ' start serial for terminal *

  start_bkg                                                     ' start pulse measure *


pub wait_for_terminal(clear, delay)

'' Wait for terminal to be open and key pressed

  term.rxflush
  term.rx
  if (clear)
    term.tx(term#CLS)
    time.pause(delay #> 0)


con { background }

var

  long  bcog
  long  bcstack[32]

  long  millis
  long  pulsems


pub start_bkg

  stop_bkg

  bcog := cognew(background, @bcstack) + 1


pub stop_bkg

  if (bcog)
    cogstop(bcog - 1)

  millis  := 0
  pulsems := 0


pri background | sample, pwidth, t

  repeat                                                        ' wait for input to be low
    if (ina[PLS_IN] == 0)
      quit
    waitcnt(cnt + MS_001)

  sample := %00                                                 ' clear locals
  pwidth := 0

  t := cnt                                                      ' sync loop timer
  repeat
    waitcnt(t += MS_001)
    millis += 1

    sample := ((sample << 1) | ina[PLS_IN]) & %11               ' sample, save last & current

    case sample
      %01 : pwidth  := 1                                        ' leading edge, reset
      %11 : pwidth  += 1                                        ' input high, accumulate
      %10 : pulsems := pwidth                                   ' falling edge, report
      %00 : pwidth  := 0                                        ' input low


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}